package mjml

import (
	"bytes"
	"context"
	"errors"
	"fmt"
	"io"

	"github.com/rs/zerolog"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/rendering"
)

var (
	ErrInvalidOutputType = errors.New("invalid output type")

	ErrInvalidInputType = errors.New("invalid input type")
)

// BufferCloser is a bytes.Buffer that implements io.ReadCloser.
type BufferCloser struct {
	*bytes.Buffer
}

// Close does nothing.
func (b BufferCloser) Close() error {
	return nil
}

// Engine implements rendering.Engine.
type Engine struct {
	ctx *Context
}

// Info returns the engine description.
func (e Engine) Info() rendering.EngineInfo {
	return rendering.EngineInfo{
		Type:        "MJML",
		Description: "Render MJML to HTML, see [https://mjml.io](https://mjml.io)",
		Transitions: []rendering.EngineTransitionInfo{
			{FromType: "text/mjml", ToType: "text/html", Priority: 1.0},
		},
		Metadata: []rendering.EngineMetadataInfo{
			{
				Name: "mjml-validation-level", Default: "soft",
				Description: `Set the mjml engine validation level.
                Can be one of 'skip', 'soft', 'strict'.
                See the [Documentation for MJML](https://mjml.io/documentation/#validating-mjml) for more information.`,
			},
		},
	}
}

func wrapErrors(errors []Error) []models.RenderError {
	if len(errors) == 0 {
		return nil
	}
	renderErrors := make([]models.RenderError, len(errors))
	for i, err := range errors {
		renderErrors[i].Engine = "MJML"
		renderErrors[i].Line = int64(err.Line)
		renderErrors[i].Message = err.Message
		renderErrors[i].Extra = make(map[string]string)
		renderErrors[i].Extra["tagname"] = err.TagName
		renderErrors[i].Extra["formatted-message"] = err.FormattedMessage
	}

	return renderErrors
}

// Render a document from a html doc to pdf.
func (e Engine) Render(
	ctx context.Context, w rendering.DocumentWriter, r rendering.DocumentReader, toType string,
) error {
	if toType != "text/html" {
		return fmt.Errorf("%w: %s", ErrInvalidOutputType, toType)
	}

	mimetype, metadata, err := r.Metadata()
	if err != nil {
		return err
	}
	if mimetype != "text/mjml" {
		return fmt.Errorf("%w: %s", ErrInvalidInputType, mimetype)
	}

	validationLevel := ValidationSoft
	var errors []models.RenderError

	if s, ok := metadata["mjml-validation-level"]; ok {
		switch s {
		case "skip":
			validationLevel = ValidationSkip
		case "soft":
			validationLevel = ValidationSoft
		case "strict":
			validationLevel = ValidationStrict
		default:
			errors = append(errors, models.RenderError{
				Engine: "MJML",
				Message: `Invalid mjml-validation-level metadata value.
                Expects 'skip', 'soft' or 'strict', got: '` + s +
					"'. Falling back to 'soft'",
			})
		}
	}
	mjmlSrc, err := io.ReadAll(r)
	if err != nil {
		return fmt.Errorf("could not read input document: %w", err)
	}

	result, err := renderDukTape(e.ctx, string(mjmlSrc), Options{
		Beautify:        true,
		ValidationLevel: validationLevel,
	})
	if err != nil {
		return err
	}

	var outputMetadata models.Metadata
	outputMetadata.Add("mjml-version", result.Version)

	w.Metadata().Update(metadata)
	w.Metadata().Update(outputMetadata)

	w.RenderErrors(r.RenderErrors()...)
	w.RenderErrors(errors...)
	w.RenderErrors(wrapErrors(result.Errors)...)
	if _, err := w.Write([]byte(result.HTML)); err != nil {
		return fmt.Errorf("could not write: %w", err)
	}

	return nil
}

// NewEngine creates a wpd rendering engine.
func NewEngine() (*Engine, error) {
	ctx, err := initDukTapeCtx()
	if err != nil {
		return nil, err
	}

	return &Engine{
		ctx: ctx,
	}, nil
}

func init() {
	rendering.RegisterRenderingEngineSetup(
		"MJML",
		func() interface{} {
			return nil
		},
		func(optionGroup interface{}, log zerolog.Logger) (rendering.Engine, error) {
			return NewEngine()
		},
	)
}
