package handlers

import (
	"context"
	"database/sql"
	"errors"
	"time"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/restapi/operations/template_management"
)

// HandleTemplateActivity deals with template activity from the renderlogs table.
func HandleTemplateActivity(
	ctx context.Context,
	db *sqlx.DB,
	tmpl *models.TemplateDB,
	log *zerolog.Logger,
) error {
	var (
		tmplID                  = int(tmpl.ID)
		username                = tmpl.Template.Account
		totalRenderCount        = 0
		currentMonthRenderCount = 0
		lastMonthRenderCount    = 0
		month                   = time.Now().Month()
		lastMonth               = time.Now().AddDate(0, -1, 0).Month()
	)

	totalQuery := squirrel.Select("COUNT(*)").
		From(models.RenderLogTable).
		Where(squirrel.Eq{models.RenderLogTemplateIDColumn: tmplID}).
		Where(squirrel.Eq{models.RenderLogUsernameColumn: username})
	if err := database.GetContext(
		ctx, db, &totalRenderCount, totalQuery, log); err != nil {
		return err
	}

	currentMonthQuery := squirrel.Select("COUNT(*)").
		From(models.RenderLogTable).
		Where(squirrel.Eq{models.RenderLogTemplateIDColumn: tmplID}).
		Where(squirrel.Eq{models.RenderLogUsernameColumn: username}).
		Where(squirrel.Expr("EXTRACT(MONTH FROM rendered_at) = ?", month))
	if err := database.GetContext(
		ctx, db, &currentMonthRenderCount, currentMonthQuery, log); err != nil {
		return err
	}

	lastMonthQuery := squirrel.Select("COUNT(*)").
		From(models.RenderLogTable).
		Where(squirrel.Eq{models.RenderLogTemplateIDColumn: tmplID}).
		Where(squirrel.Eq{models.RenderLogUsernameColumn: username}).
		Where(squirrel.Expr("EXTRACT(MONTH FROM rendered_at) = ?", lastMonth))
	if err := database.GetContext(
		ctx, db, &lastMonthRenderCount, lastMonthQuery, log); err != nil {
		return err
	}
	activity := models.Activity{
		RenderStats: &models.RenderStats{
			Total:        int64(totalRenderCount),
			CurrentMonth: int64(currentMonthRenderCount),
			LastMonth:    int64(lastMonthRenderCount),
		},
	}
	tmpl.Template.Activity = &activity

	return nil
}

// NewAccountTemplateGetHandler returns a AccountTemplateGetHandler.
func NewAccountTemplateGetHandler(db *sqlx.DB) *AccountTemplateGetHandler {
	return &AccountTemplateGetHandler{
		db: db,
	}
}

// AccountTemplateGetHandler handles GET /user.
type AccountTemplateGetHandler struct {
	db *sqlx.DB
}

func (h *AccountTemplateGetHandler) handle(
	ctx context.Context, account, name string,
) (*models.Template, error) {
	log := zerolog.Ctx(ctx)

	var templatedb models.TemplateDB

	db := database.NewSQLHelper(ctx, h.db, *log)

	if err := db.GetWhere(&templatedb, squirrel.Eq{
		models.TemplateDBAccountColumn: account,
		models.TemplateDBNameColumn:    name,
	}); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			return nil, NotFoundf("no such template: %s/%s", account, name)
		}

		return nil, err
	}
	templatedb.FromDB()
	templatedb.ForceBase64()
	if err := HandleTemplateActivity(ctx, h.db, &templatedb, log); err != nil {
		return nil, err
	}

	return &templatedb.Template, nil
}

// Handle handles the incoming request.
func (h *AccountTemplateGetHandler) Handle(
	params template_management.AccountTemplateGetParams, principal *models.Principal,
) middleware.Responder {
	ctx := params.HTTPRequest.Context()
	log := zerolog.Ctx(ctx)
	if principal.IsAdmin {
		log.Error().Msg("admin cannot manage templates. Please use a non-admin user")

		return template_management.
			NewAccountTemplateAddUnauthorized()
	}
	if principal.Subject != params.Account {
		log.Error().Msg("user can manage its own templates only")

		return template_management.
			NewAccountTemplateAddUnauthorized()
	}
	template, err := h.handle(params.HTTPRequest.Context(), params.Account, params.Name)
	if err != nil {
		return HandleError(err, log)
	}

	return template_management.NewAccountTemplateGetOK().WithPayload(template)
}
