package handlers

import (
	"context"
	"database/sql"
	"errors"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	redner "orus.io/orus-io/rednerd/lib"
	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/rendering"
	op_auth "orus.io/orus-io/rednerd/restapi/operations/auth"
)

func NewForgotPassword(
	db *sqlx.DB,
	registry *rendering.EngineRegistry,
	templateRegistry *rendering.TemplateEngineRegistry,
	baseURL string,
) *ForgotPassword {
	return &ForgotPassword{
		db:               db,
		registry:         registry,
		templateRegistry: templateRegistry,
		baseURL:          baseURL,
	}
}

type ForgotPassword struct {
	db               *sqlx.DB
	registry         *rendering.EngineRegistry
	templateRegistry *rendering.TemplateEngineRegistry
	baseURL          string
}

func (h *ForgotPassword) handle(
	ctx context.Context, username, email string,
) error {
	log := zerolog.Ctx(ctx)
	tx, err := database.Begin(ctx, h.db)
	if err != nil {
		return err
	}
	defer tx.RollbackIfOpened(*log)

	db := database.NewSQLHelper(ctx, tx, *log)

	filter := squirrel.Eq{}

	if username != "" {
		filter["username"] = username
	}
	if email != "" {
		filter["email"] = email
	}
	var userDB models.UserDB
	if err := db.Get(&userDB, squirrel.Select("*").From("users").Where(filter)); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			// The user does not exist, we stop here
			return nil
		}

		return err
	}

	if userDB.Email == "" {
		// We have no email for this account, stopping here
		return nil
	}

	var (
		templateDB models.TemplateDB
		template   *models.Template
	)
	if err := db.Get(&templateDB,
		squirrel.Select(models.TemplateDBColumns...).
			From(models.TemplateDBTable).
			Where(squirrel.Eq{
				models.TemplateDBAccountColumn: "templates",
				models.TemplateDBNameColumn:    "forgot-password",
			}),
	); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			template = &redner.DefaultResetMailTemplate
		} else {
			return err
		}
	} else {
		templateDB.FromDB()
		template = &templateDB.Template
	}

	if err := redner.ResetPassword(
		ctx, db, h.templateRegistry, h.registry, h.baseURL,
		userDB, template,
	); err != nil {
		return err
	}

	return tx.Commit()
}

func (h *ForgotPassword) Handle(
	params op_auth.ForgotPasswordParams,
) middleware.Responder {
	ctx := params.HTTPRequest.Context()
	if params.Info.Username == "" && params.Info.Email == "" {
		return op_auth.NewForgotPasswordDefault(400).
			WithPayload(&models.Error{Message: "must set 'username' or 'email'"})
	}

	if err := h.handle(ctx, params.Info.Username, params.Info.Email); err != nil {
		var mErr *models.Error
		if errors.As(err, &mErr) {
			return op_auth.NewForgotPasswordDefault(400).
				WithPayload(mErr)
		}

		return InternalError(err, zerolog.Ctx(ctx), false)
	}

	return op_auth.NewForgotPasswordNoContent()
}
