package handlers

import (
	"context"
	"database/sql"
	"errors"
	"net/http"
	"time"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/go-openapi/strfmt"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	redner "orus.io/orus-io/rednerd/lib"
	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/rendering"
	op_auth "orus.io/orus-io/rednerd/restapi/operations/auth"
)

func NewAuthRegister(
	db *sqlx.DB,
	registry *rendering.EngineRegistry,
	templateRegistry *rendering.TemplateEngineRegistry,
	baseURL string,
) *AuthRegister {
	return &AuthRegister{
		db:               db,
		registry:         registry,
		templateRegistry: templateRegistry,
		baseURL:          baseURL,
	}
}

type AuthRegister struct {
	db               *sqlx.DB
	registry         *rendering.EngineRegistry
	templateRegistry *rendering.TemplateEngineRegistry
	baseURL          string
}

func (h *AuthRegister) handle(ctx context.Context, registration models.Registration) error {
	if registration.Email == "" {
		return &models.Error{Message: "email is mandatory"}
	}
	info := models.Userinfo{
		Username:  registration.Email,
		Email:     registration.Email,
		Firstname: registration.Firstname,
		Lastname:  registration.Lastname,
		CreatedAt: strfmt.DateTime(time.Now().UTC()),
		IsAdmin:   false,
	}

	log := zerolog.Ctx(ctx)
	tx, err := database.Begin(ctx, h.db)
	if err != nil {
		return err
	}
	defer tx.RollbackIfOpened(*log)

	db := database.NewSQLHelper(ctx, tx, *log)

	var exists bool
	if err := db.Get(&exists,
		squirrel.Select("1").From("users").
			Where(squirrel.Eq{models.UserDBEmailColumn: info.Email}).
			Prefix("SELECT EXISTS(").
			Suffix(")"),
	); err != nil {
		return err
	}

	if exists {
		// return &models.Error{Message: "duplicate email"}
		// TODO send password-reset anyway ?
		return nil
	}

	userDB := models.UserDB{
		Userinfo: info,
	}

	if _, err := db.Insert(&models.AccountDB{
		Name:   userDB.Username,
		IsUser: true,
	}); err != nil {
		return err
	}

	if _, err := db.Insert(&userDB); err != nil {
		return err
	}

	var (
		templateDB models.TemplateDB
		template   *models.Template
	)
	if err := db.Get(&templateDB,
		squirrel.Select(models.TemplateDBColumns...).
			From(models.TemplateDBTable).
			Where(squirrel.Eq{
				models.TemplateDBAccountColumn: "templates",
				models.TemplateDBNameColumn:    "register",
			}),
	); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			template = &redner.DefaultSubscribeTempate
		} else {
			return err
		}
	} else {
		templateDB.FromDB()
		template = &templateDB.Template
	}

	if err := redner.ResetPassword(
		ctx, db, h.templateRegistry, h.registry, h.baseURL,
		userDB, template,
	); err != nil {
		return err
	}

	return tx.Commit()
}

func (h *AuthRegister) Handle(
	params op_auth.RegisterParams,
) middleware.Responder {
	ctx := params.HTTPRequest.Context()
	if params.Registration == nil {
		return op_auth.NewRegisterDefault(http.StatusBadRequest).
			WithPayload(&models.Error{Message: "Missing body"})
	}
	err := h.handle(ctx, *params.Registration)
	var genErr *models.Error
	switch {
	case errors.As(err, &genErr):
		return op_auth.NewRegisterDefault(http.StatusBadRequest).
			WithPayload(genErr)
	case err != nil:
		return InternalError(err, zerolog.Ctx(ctx), false)
	}

	return op_auth.NewRegisterNoContent()
}
