package handlers

import (
	"context"
	"strings"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/restapi/operations/user_management"
	"orus.io/orus-io/rednerd/utils"
)

// NewUserApikeyAddHandler returns a UserApikeyAddHandler.
func NewUserApikeyAddHandler(db *sqlx.DB) *UserApikeyAddHandler {
	return &UserApikeyAddHandler{
		db: db,
	}
}

// UserApikeyAddHandler handles GET /user.
type UserApikeyAddHandler struct {
	db *sqlx.DB
}

func (h *UserApikeyAddHandler) handle(
	ctx context.Context, username string, apikey *models.Apikey,
) (*models.ApikeyAndValue, error) {
	var (
		log      = zerolog.Ctx(ctx)
		apikeydb = models.ApikeyDB{
			ApikeyAndValue: models.ApikeyAndValue{
				Name:        apikey.Name,
				Description: apikey.Description,
			},
			Username: username,
		}
	)
	apikeydb.Value = utils.RandomString(32)

	columns := apikeydb.Columns(false)
	q := squirrel.Insert(apikeydb.Table()).
		Columns(columns...)
	q = q.Values(apikeydb.Values(columns...)...)

	q = q.Suffix("RETURNING " + strings.Join(models.ApikeyColumns, ","))

	tx, err := database.Begin(ctx, h.db)
	if err != nil {
		return nil, err
	}
	defer tx.RollbackIfOpened(*log)

	if err := database.GetContext(ctx, tx, &apikeydb, q, log); err != nil {
		return nil, err
	}
	if err := tx.Commit(); err != nil {
		return nil, err
	}

	return &apikeydb.ApikeyAndValue, nil
}

// Handle handles the incoming request.
func (h *UserApikeyAddHandler) Handle(
	params user_management.UserApikeyAddParams, principal *models.Principal,
) middleware.Responder {
	ctx := params.HTTPRequest.Context()
	log := zerolog.Ctx(ctx)
	if !principal.IsAdmin && principal.Subject != params.Username {
		log.Error().Msg("only admins can add user api keys, expect if you are adding your own key")

		return user_management.NewUserApikeyAddUnauthorized()
	}
	apikey, err := h.handle(params.HTTPRequest.Context(), params.Username, params.Apikey)
	if err != nil {
		return InternalError(err, log, false)
	}

	return user_management.NewUserApikeyAddCreated().WithPayload(apikey)
}
